package be.iminds.ilabt.jfed.rspec.model.imutable_impl;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecWriter;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.text.ParseException;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * ImmutableModelRspec
 */
public class ImmutableModelRspec implements ModelRspec {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ImmutableModelRspec.class);

    private final List<ExtraXml> extraXml;
    private final Map<String, String> origNameSpaceInfo;

    private final List<ImmutableRspecNode> nodes;
    private final List<ImmutableRspecLink> links;
    private final List<ImmutableLease> leases;
    private final List<ImmutableChannel> channels;
    private final List<ImmutableGeantTestbedType> geantTestbedTypes;
    private final List<ImmutableOpenflowDataPath> openflowDataPaths;
    private final List<OpenflowSliver> openflowSlivers;
    private final List<ImmutableAddressPool> addressPools;
    private final List<ImmutableExecuteAnsiblePlaybook> executeAnsiblePlaybooks;
    private final List<ImmutableDistributeSshKeypair> distributeSshKeypairs;
    private final String origDefaultNamespace;
    private final String origSchemaLocation;
    private final String rspecType;
    private final List<UserSpec> sshKeys;
    private final String expiresString;
    private final HardwareTypeInfo hardwareTypeInfo;

    private final Integer routableAddressesAvailable;
    private final Integer routableAddressesConfigured;
    private final Integer multiplexFactor;

    private final RspecFactory factory;

    private static <T> List<T> listInit(List<T> orig) {
        if (orig == null) return null;
        return Collections.unmodifiableList(new ArrayList<>(orig));
    }
    private static <OT, NT> List<NT> listInit(List<OT> orig, Function<OT, NT> convert) {
        if (orig == null) return null;
        return Collections.unmodifiableList(orig.stream().map(convert).collect(Collectors.toList()));
    }
    private static <T, Y> Map<T, Y> mapInit(Map<T, Y> orig) {
        if (orig == null) return null;
        return Collections.unmodifiableMap(new TreeMap<>(orig));
    }

    //package private, use ImmutableRSpecFactory to copy
    ImmutableModelRspec(ImmutableRspecFactory factory, ModelRspec orig) {
        this.factory = factory;

        this.rspecType = orig.getType();
        this.extraXml = listInit(orig.getExtraXml());
        this.origNameSpaceInfo = mapInit(orig.getNameSpaceInfo());

        this.leases = listInit(orig.getLeases(), ImmutableLease::new);
        this.channels = listInit(orig.getChannels(), ImmutableChannel::new);
        this.geantTestbedTypes = listInit(orig.getGeantTestbedTypes(), ImmutableGeantTestbedType::new);
        this.openflowDataPaths = listInit(orig.getOpenflowDataPaths(), ImmutableOpenflowDataPath::new);
        this.openflowSlivers = listInit(orig.getOpenflowSlivers(), OpenflowSliver::new);
        this.addressPools = listInit(orig.getAddressPools(), ImmutableAddressPool::new);
        this.executeAnsiblePlaybooks = listInit(orig.getExecuteAnsiblePlaybooks(), ImmutableExecuteAnsiblePlaybook::new);
        this.distributeSshKeypairs = listInit(orig.getDistributeSshKeypairs(), ImmutableDistributeSshKeypair::new);
        this.sshKeys = listInit(orig.getSshKeys());

        this.origDefaultNamespace = orig.getOrigDefaultNamespace();
        this.origSchemaLocation = orig.getSchemaLocation();
        this.expiresString = orig.getExpiresString();
        this.hardwareTypeInfo = orig.getHardwareTypeInfo();
        this.routableAddressesAvailable = orig.getRoutableAddressesAvailable();
        this.routableAddressesConfigured = orig.getRoutableAddressesConfigured();
        this.multiplexFactor = orig.getMultiplexFactor();

        this.nodes = listInit(orig.getNodes(), origNode -> new ImmutableRspecNode(this, origNode));
        this.links = listInit(orig.getLinks(), origLink -> new ImmutableRspecLink(this, origLink));

        //make all interface immutable
        nodes.stream().flatMap(node -> node.getInterfaces().stream()).forEach(iface -> iface.setForceFinal(true));
        links.stream().flatMap(node -> node.getInterfaces().stream()).forEach(iface -> iface.setForceFinal(true));
    }

    @Override
    @Nonnull
    public ModelRspecType getModelRspecType() {
        return ModelRspecType.IMMUTABLE;
    }

    @Override
    public void setParseData(@Nullable String schemaLocation, @Nullable String defaultNamespace, @Nonnull Map<String, String> namespaces) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public String getSchemaLocation() {
        return origSchemaLocation;
    }

    @Override
    @Nullable
    public String getOrigDefaultNamespace() {
        return origDefaultNamespace;
    }

    @Override
    @Nonnull
    public Map<String, String> getNameSpaceInfo() {
        return origNameSpaceInfo;
    }

    @Override
    public void addNode(@Nonnull RspecNode node) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void addLink(@Nonnull RspecLink link) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<ImmutableRspecNode> getNodes() {
        return nodes;
    }

    @Override
    @Nonnull
    public List<ImmutableRspecLink> getLinks() {
        return links;
    }

    @Override
    public void addOpenflowDataPath(@Nonnull OpenflowDataPath dp) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<ImmutableOpenflowDataPath> getOpenflowDataPaths() {
        return openflowDataPaths;
    }

    @Override
    public void addOpenflowSliver(@Nonnull OpenflowSliver sliver) {
        throw new UnsupportedOperationException();
    }

    @Nonnull
    @Override
    public List<? extends OpenflowSliver> getOpenflowSlivers() {
        return openflowSlivers;
    }


    @Override
    public void addLease(@Nonnull Lease lease) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<? extends Lease> getLeases() {
        return leases;
    }

    @Override
    public void addChannel(@Nonnull Channel channel) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<ImmutableChannel> getChannels() {
        return channels;
    }

    @Override
    public void addGeantTestbedType(@Nonnull GeantTestbedType gtt) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<? extends GeantTestbedType> getGeantTestbedTypes() {
        return geantTestbedTypes;
    }

    @Override
    public void addAddressPool(@Nonnull AddressPool addressPool) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean deleteAddressPool(@Nonnull AddressPool addressPool) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<ImmutableAddressPool> getAddressPools() {
        return addressPools;
    }

    @Override
    @Nullable
    public AddressPool getAddressPoolByClientId(@Nonnull String clientId) {
        for (ImmutableAddressPool addressPool : addressPools) {
            if (Objects.equals(clientId, addressPool.getClientId()))
                return addressPool;
        }
        return null;
    }

    @Override
    public void addExecuteAnsiblePlaybook(@Nonnull ExecuteAnsiblePlaybook executeAnsiblePlaybook) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean deleteExecuteAnsiblePlaybook(@Nonnull ExecuteAnsiblePlaybook executeAnsiblePlaybook) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<? extends ImmutableExecuteAnsiblePlaybook> getExecuteAnsiblePlaybooks() {
        return executeAnsiblePlaybooks;
    }


    @Override
    public void addDistributeSshKeypair(@Nonnull DistributeSshKeypair distributeSshKeypair) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean deleteDistributeSshKeypair(@Nonnull DistributeSshKeypair distributeSshKeypair) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<? extends DistributeSshKeypair> getDistributeSshKeypairs() {
        return distributeSshKeypairs;
    }

    @Nullable
    @Override
    public HardwareTypeInfo getHardwareTypeInfo() {
        return hardwareTypeInfo;
    }

    @Override
    public void setHardwareTypeInfo(@Nullable HardwareTypeInfo hardwareTypeInfo) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public ImmutableRspecNode getNodeByUniqueId(@Nonnull String id) {
        for (ImmutableRspecNode node : nodes)
            if (Objects.equals(node.getUniqueId(), id))
                return node;
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecNode getNodeByClientId(@Nonnull String id) {
        for (ImmutableRspecNode node : nodes)
            if (Objects.equals(node.getClientId(), id))
                return node;
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecNode getNodeByComponentId(@Nonnull String id) {
        for (ImmutableRspecNode node : nodes)
            if (Objects.equals(node.getComponentId().toString(), id))
                return node;
        return null;
    }

    @Override
    @Nullable
    public ImmutableChannel getChannelByComponentId(@Nonnull GeniUrn urn) {
        for (ImmutableChannel channel : channels)
            if (Objects.equals(channel.getComponentId(), urn))
                return channel;
        return null;
    }


    @Override
    @Nonnull
    public List<ImmutableRspecNode> getNodesByAuthority(@Nonnull Server server) {
        return nodes.stream()
                .filter(node -> node.getComponentManagerId().equalsIgnoringSubAuth(server.getDefaultComponentManagerUrn()))
                .collect(Collectors.toList());
    }

    @Override
    @Nullable
    public ImmutableRspecLink getLinkByClientId(@Nonnull String id) {
        for (ImmutableRspecLink link : links)
            if (Objects.equals(link.getClientId(), id))
                return link;
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecLink getLinkByUniqueId(@Nonnull String id) {
        for (ImmutableRspecLink link : links)
            if (Objects.equals(link.getUniqueId(), id))
                return link;
        return null;
    }

    @Override
    public void deleteNode(@Nonnull RspecNode node) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void deleteLink(@Nonnull RspecLink link) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean deleteChannel(@Nonnull Channel channel) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public String toGeni3Rspec(@Nonnull RequestRspecSpecialCases specialCase) {
        StaxRspecWriter staxRspecWriter = new StaxRspecWriter(this);

        switch (specialCase) {
            case NONE: {
                //default staxRspecWriter options are fine
                break;
            }
            case PLE: {
                //already using quick hack to get ple to work in: RspecNode.writePropertiesToGeni3RequestRspec()
                List<String> allowedComponentManagers = new ArrayList<>();
                allowedComponentManagers.add("urn:publicid:IDN+ple:ibbtple+authority+cm");
                //hack for ple: don't add nodes from other componentmanagers
                staxRspecWriter.setComponentManagerFilter(allowedComponentManagers);

                //another hack for ple: ple also doesn't like schemaLocation
                staxRspecWriter.setAddSchemaLocation(false);

                //yet another hack for PLE: add a meaningless, empty <sliver/> element to each node
                staxRspecWriter.setAddEmptySliverToNodeHack(true);
                break;
            }
            case NITOS_NO_E_NODE_B_HACK: {
                staxRspecWriter.setUseNitosENodeBHack(false);
                break;
            }
            default:
                throw new RuntimeException("Special case " + specialCase + " not supported");
        }

        try {
            return staxRspecWriter.call();
        } catch (RspecParseException e) {
            LOG.error("Error converting to Rspec: ", e);
            throw new RuntimeException("Error converting to Rspec: " + e, e);
        }
    }

    @Override
    @Nonnull
    public String toGeni3Rspec() {
        return toGeni3Rspec(RequestRspecSpecialCases.NONE);
    }

    @Override
    @Nullable
    public Date getExpires() throws ParseException {
        if (expiresString == null) return null;
        return RFC3339Util.rfc3339StringToDate(expiresString);
    }

    @Override
    public void setExpires(@Nullable Date expires) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public String getExpiresString() {
        return expiresString;
    }

    @Override
    public void setExpires(@Nullable String expires) {
        throw new UnsupportedOperationException();
    }


    @Override
    @Nullable
    public Integer getRoutableAddressesAvailable() {
        return routableAddressesAvailable;
    }

    @Override
    @Nullable
    public Integer getRoutableAddressesConfigured() {
        return routableAddressesConfigured;
    }

    @Override
    public void setRoutableAddressesAvailable(@Nullable Integer routableAddressesAvailable) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setRoutableAddressesConfigured(@Nullable Integer routableAddressesConfigured) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nullable
    public GeantTestbedType getGeantTestbedTypeByName(@Nonnull String geantTestbedTypeName) {
        for (ImmutableGeantTestbedType geantTestbedType : geantTestbedTypes) {
            if (Objects.equals(geantTestbedType.getName(), geantTestbedTypeName))
                return geantTestbedType;
        }
        return null;
    }

    @Override
    @Nonnull
    public List<GeniUrn> getAllComponentManagerUrns() {
        Set<GeniUrn> urns = new HashSet<>();

        nodes.stream().map(RspecNode::getComponentManagerId).filter(Objects::nonNull).forEach(urns::add);
        links.stream().flatMap(link -> link.getComponentManagerUrns().stream()).forEach(urns::add);
        channels.stream().map(Channel::getComponentManagerId).forEach(urns::add);
        addressPools.stream().map(AddressPool::getComponentManagerId).forEach(urns::add);
        geantTestbedTypes.stream().map(GeantTestbedType::getComponentManagerId).forEach(urns::add);

        return new ArrayList<>(urns);
    }

    @Override
    @Nonnull
    public List<UserSpec> getSshKeys() {
        return sshKeys;
    }

    @Override
    public void setSshKeys(@Nonnull List<UserSpec> newKeys) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public String nextLinkName() {
        int c = links.size();
        String res = "link" + c;
        while (getLinkByClientId(res) != null)
            res = "link" + (++c);

        return res;
    }

    @Override
    @Nonnull
    public String nextNodeName() {
        int c = nodes.size();
        String res = "node" + c;
        while (getNodeByClientId(res) != null)
            res = "node" + (++c);

        return res;
    }

    @Override
    @Nullable
    public String getType() {
        return rspecType;
    }

    @Override
    @Nonnull
    public String nextIfaceName(@Nonnull RspecNode node) {
        int c = node.getInterfaces().size();
        String res = node.getClientId() + ":if" + c;
        while (getInterfaceByClientId(res) != null)
            res = node.getClientId() + ":if" + (++c);

        return res;
    }

    @Override
    @Nullable
    public ImmutableRspecInterface getInterfaceByClientId(@Nonnull String clientId) {
        if (nodes != null) {
            for (ImmutableRspecNode node : nodes) {
                ImmutableRspecInterface iface = node.getInterfaceByClientId(clientId);
                if (iface != null)
                    return iface;
            }
        }
        if (links != null) {
            for (ImmutableRspecLink link : links) {
                ImmutableRspecInterface iface = link.getInterfaceByClientId(clientId);
                if (iface != null)
                    return iface;
            }
        }
        return null;
    }

    @Override
    @Nullable
    public ImmutableRspecInterface getInterfaceByUniqueId(@Nonnull String uniqueId) {
        if (nodes != null) {
            for (ImmutableRspecNode node : nodes) {
                ImmutableRspecInterface iface = node.getInterfaceByUniqueId(uniqueId);
                if (iface != null)
                    return iface;
            }
        }
        if (links != null) {
            for (ImmutableRspecLink link : links) {
                ImmutableRspecInterface iface = link.getInterfaceByUniqueId(uniqueId);
                if (iface != null)
                    return iface;
            }
        }
        return null;
    }

    @Override
    @Nullable
    public Integer getMultiplexFactor() {
        return multiplexFactor;
    }

    @Override
    public void setMultiplexFactor(@Nullable Integer multiplexFactor) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public List<ExtraXml> getExtraXml() {
        return extraXml;
    }

    @Override
    public void setExtraXml(@Nonnull List<ExtraXml> extraXml) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public ModelRspec copy() {
        return factory.copyModelRspec(this, null/*new rspec type (request/manifest/...)*/);
    }
}
